/**
  *  \file client/actions/gabuild.cc
  *  \brief Global Actions: Building
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions/gabuild.h"
#include "client/dialogs/goaldialog.h"
#include "cpluslib/format.h"
#include "game/actions/planetstructs.h"
#include "game/planet.h"
#include "game/transact.h"
#include "int/mutex.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/planettransfer.h"

/************************* WGlobalAutobuildAction ************************/

class WGlobalAutobuildAction::Executer : public WGlobalActionExecuter {
 public:
    bool execute(GObject& obj, bool lock);
    void finish(int obj, int success);
};

WGlobalActionExecuter*
WGlobalAutobuildAction::prepare()
{
    if (UIMessageBox(_("Perform auto-build on all planets?"), _("Global Actions")).doYesNoDialog()) {
        return new Executer();
    } else {
        return 0;
    }
}

bool
WGlobalAutobuildAction::Executer::execute(GObject& obj, bool lock)
{
    bool did = false;
    GPlanet* pl = dynamic_cast<GPlanet*>(&obj);
    if (pl != 0 && pl->isPlayable(GObject::Playable)) {
        if (lock || IntMutex::query(format("P%d.BUILD", pl->getId())) == 0) {
            GMultiTransaction mt;
            GPlanetTransfer& transfer         = mt.addNewSubtransaction(new GPlanetTransfer(*pl));
            GPlanetBuildStructuresAction& bsa = mt.addNewSubtransaction(new GPlanetBuildStructuresAction(*pl, transfer));
            bsa.doStandardAutoBuild();
            if (!bsa.getCost().isEmpty()) {
                did = true;
            }
            mt.doCommit();
        }
    }
    return did;
}

void
WGlobalAutobuildAction::Executer::finish(int /*obj*/, int success)
{
    messageBox(format(_("Structures have been built on %d planet%!1{s%}."), success), _("Global Actions"));
}

/*********************** WGlobalSetBuildGoalsAction **********************/

class WGlobalSetBuildGoalsAction::Executer : public WGlobalActionExecuter {
    int goals[4];
    int speeds[4];
 public:
    Executer(const WBuildGoalsDialog& dlg);
    bool execute(GObject& obj, bool /*lock*/);
    void finish(int obj, int success);
};

WGlobalActionExecuter*
WGlobalSetBuildGoalsAction::prepare()
{
    WBuildGoalsDialog dlg(true);
    ui_root->centerWidget(dlg);
    if (ui_root->runModal(dlg)) {
        return new Executer(dlg);
    } else {
        return 0;
    }
}

WGlobalSetBuildGoalsAction::Executer::Executer(const WBuildGoalsDialog& dlg)
{
    for (int i = 0; i < 4; ++i) {
        goals[i] = dlg.getGoal(TPlanetaryStructure(i));
        speeds[i] = dlg.getSpeed(TPlanetaryStructure(i));
    }
}

bool
WGlobalSetBuildGoalsAction::Executer::execute(GObject& obj, bool /*lock*/)
{
    bool did = false;
    GPlanet* pl = dynamic_cast<GPlanet*>(&obj);
    if (pl != 0) {
        /* This action can also operate on unplayed planets! */
        for (int i = 0; i < 4; ++i) {
            TPlanetaryStructure ps = TPlanetaryStructure(i);
            if (goals[i] >= 0 && goals[i] != pl->getAutobuildGoal(ps)) {
                pl->setAutobuildGoal(ps, goals[i]);
                did = true;
            }
            if (speeds[i] >= 0 && speeds[i] != pl->getAutobuildSpeed(ps)) {
                pl->setAutobuildSpeed(ps, speeds[i]);
                did = true;
            }
        }
    }
    return did;
}

void
WGlobalSetBuildGoalsAction::Executer::finish(int /*obj*/, int success)
{
    messageBox(format(_("Auto-build goals have been changed on %d planet%!1{s%}."), success), _("Global Actions"));
}
