/**
  *  \file client/dialogs/shipcostcalc.cc
  *  \brief Ship Cost Calculator
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/dialogs/shipcostcalc.h"
#include "cpluslib/misc.h"
#include "game/planet.h"
#include "client/dialogs/shipbuilddlg.h"
#include "game/spec.h"
#include "ui/layout.h"
#include "ui/widgets/button.h"
#include "client/widgets/helpwidget.h"
#include "util/nls.h"
#include "ui/commands.h"
#include "game/team.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/quit.h"

namespace {
    /** Ship cost calculator dialog. This adds the cost calculator
        controls to a WShipBuildDialog. */
    class WShipCostCalcDialog : public WShipBuildDialog {
        UIButton* btn_tech;
        UIButton* btn_storage;
     public:
        WShipCostCalcDialog(GPlanet& p, std::vector<int>& hulls, bool storage);
        void init(bool storage);
        void onToggleTechUpgrade();
        void onToggleUseParts();
        void updateButtons();
    };
}

/** Constructor.
    \param p Planet to work on (must have a starbase and be played)
    \param hulls Hulls to offer (should be list of all hulls)
    \param storage true to offer use of starbase storage */
WShipCostCalcDialog::WShipCostCalcDialog(GPlanet& p, std::vector<int>& hulls, bool storage)
    : WShipBuildDialog(_("Starship Cost Calculator"), p, false, hulls)
{
    init(storage);
}

void
WShipCostCalcDialog::init(bool storage)
{
    // HBox
    //   [x] Use Tech Levels   [x] Use Parts from Storage
    // HBox
    //   UIButton "OK"
    //   UIButton "D-Detailed Bill"
    //   UISpacer
    //   UIButton "Help"

    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    btn_tech = &h.add(new UIButton(9001, "I", 'i'));
    g1.add(*btn_tech);
    g1.add(h.add(new UIStaticText(_("Include tech upgrades"), UIStaticText::WhiteTitle)));
    if (storage) {
        btn_storage = &h.add(new UIButton(9002, "U", 'u'));
        g1.add(*btn_storage);
        g1.add(h.add(new UIStaticText(_("Use parts from storage"), UIStaticText::WhiteTitle)));
    } else {
        btn_storage = 0;
    }
    g1.add(h.add(new UISpacer()));
    add(g1);

    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIButton& btn_bill = h.add(new UIButton(9000, _("D - Detailed Bill"), 'd'));
    g2.add(h.add(new UIButton(cm_Enter, _("Close"), SDLK_ESCAPE)).addStop());
    g2.add(btn_bill);
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(_("Help"), 'h')));
    add(g2);
    add(h.add(new WHelpWidget("pcc2:buildship")));
    add(h.add(new UIQuit(cm_Enter)));

    btn_bill.sig_fired.add(static_cast<WShipBuildDialog*>(this), &WShipBuildDialog::onDetailedBill);
    btn_tech->sig_fired.add(this, &WShipCostCalcDialog::onToggleTechUpgrade);
    if (btn_storage)
        btn_storage->sig_fired.add(this, &WShipCostCalcDialog::onToggleUseParts);

    action.setUseTechUpgrade(true);
    action.setUsePartsFromStorage(false);

    pack();
    updateButtons();
}

void
WShipCostCalcDialog::onToggleTechUpgrade()
{
    action.setUseTechUpgrade(!action.isUseTechUpgrade());
    updateButtons();
    onPlanetChanged(); // FIXME: see baseshipbuild.cc; should use signals properly.
}

void
WShipCostCalcDialog::onToggleUseParts()
{
    action.setUsePartsFromStorage(!action.isUsePartsFromStorage());
    updateButtons();
    onPlanetChanged(); // FIXME: see baseshipbuild.cc; should use signals properly.
}

void
WShipCostCalcDialog::updateButtons()
{
    btn_tech->setState(UIButton::bf_Highlight, action.isUseTechUpgrade());
    if (btn_storage)
        btn_storage->setState(UIButton::bf_Highlight, action.isUsePartsFromStorage());
}

static bool
sortByHullName(int a, int b)
{
    return strUCase(getHull(a).getName()) < strUCase(getHull(b).getName());
}

/** Ship cost calculator, internal main entry point.
    \param p fully-prepared planet (with starbase)
    \param storage true to offer build from storage */
static void
doSCC(GPlanet& p, bool storage)
{
    // Prepare hull list
    std::vector<int> hulls;
    for (int i = 1, e = getNumHulls(); i <= e; ++i) {
        hulls.push_back(i);
    }
    std::sort(hulls.begin(), hulls.end(), sortByHullName);

    // Prepare dialog
    WShipCostCalcDialog dlg(p, hulls, storage);
    ui_root->centerWidget(dlg);
    ui_root->runModal(dlg);
}

/****************************** Preparation ******************************/

/** Fill all unknown fields of a planet with default values. */
static void
fillUnknownFields(GPlanet& p)
{
    // Owner and colony
    if (!p.isOwnerKnown() || p.getOwner() == 0)
        p.setOwner(mp16_t(getPlayerId()));
    if (!p.getColonistHappiness().isKnown())
        p.setColonistHappiness(mn16_t(100));
    if (!p.getColonistTax().isKnown())
        p.setColonistTax(mp16_t(0));
    if (!p.getColonists().isKnown() || p.getColonists() == 0)
        p.setColonists(mn32_t(100));

    // Structures
    if (!p.getStructures(str_Mines).isKnown())
        p.setStructures(str_Mines, mp16_t(0));
    if (!p.getStructures(str_Factories).isKnown())
        p.setStructures(str_Factories, mp16_t(0));
    if (!p.getStructures(str_Defense).isKnown())
        p.setStructures(str_Defense, mp16_t(0));

    // Natives
    if (!p.getNatives().isKnown() || !p.getNativeRace().isKnown() || p.getNatives() == 0) {
        p.setNatives(mp32_t(0));
        p.setNativeRace(mp16_t(0));
        p.setNativeGovernment(mp16_t(0));
        p.setNativeHappiness(mn16_t(100));
        p.setNativeTax(mp16_t(0));
        p.setKnownToHaveNatives(false);
    } else {
        if (!p.getNativeGovernment().isKnown())
            p.setNativeGovernment(mp16_t(5));
        if (!p.getNativeHappiness().isKnown())
            p.setNativeHappiness(mn16_t(100));
        if (!p.getNativeTax().isKnown())
            p.setNativeTax(mp16_t(0));
    }

    // FCode
    if (!p.isFCodeKnown())
        p.setFCode("hhg");

    // Base. A default base will be created later
    p.setBuildBaseFlag(false);

    // Environment
    if (!p.getTemperature().isKnown())
        p.setTemperature(mp16_t(50));

    for (int i = el_Neutronium; i <= el_Molybdenum; ++i) {
        if (!p.getOreDensity(i).isKnown())
            p.setOreDensity(i, mp16_t(50));
        if (!p.getOreGround(i).isKnown())
            p.setOreGround(i, mp32_t(1000));
        if (!p.getCargoRaw(GCargoType(i)).isKnown())
            p.setCargoRaw(GCargoType(i), mp32_t(1000));
    }

    // Cargo
    if (!p.getCargoRaw(el_Supplies).isKnown())
        p.setCargoRaw(el_Supplies, mp32_t(10000));
    if (!p.getCargoRaw(el_Money).isKnown())
        p.setCargoRaw(el_Money, mp32_t(10000));
}

/** Create a starbase at a planet. */
static void
createDefaultBase(GPlanet& p)
{
    TStarbase sb;
    sb.base_id          = p.getId();
    sb.owner            = p.getOwner();
    sb.defense          = 0;
    sb.damage           = 0;
    std::fill_n(sb.tech_levels,    countof(sb.tech_levels),    1);
    std::fill_n(sb.engine_store,   countof(sb.engine_store),   0);
    std::fill_n(sb.hull_store,     countof(sb.hull_store),     0);
    std::fill_n(sb.beam_store,     countof(sb.beam_store),     0);
    std::fill_n(sb.launcher_store, countof(sb.launcher_store), 0);
    std::fill_n(sb.ammo_store,     countof(sb.ammo_store),     0);
    sb.shipyard_id      = 0;
    sb.shipyard_action  = 0;
    sb.mission          = 0;
    sb.build.hull       = 0;
    sb.build.engine     = 0;
    sb.build.beam_type  = 0;
    sb.build.beam_count = 0;
    sb.build.torp_type  = 0;
    sb.build.torp_count = 0;

    switch (p.getNativeRace()) {
     case nr_Ghipsoldal:
        sb.tech_levels[tl_Engine] = 10;
        break;
     case nr_Humanoid:
        sb.tech_levels[tl_Hull] = 10;
        break;
     case nr_Amphibian:
        sb.tech_levels[tl_Beam] = 10;
        break;
     case nr_Siliconoid:
        sb.tech_levels[tl_Torpedo] = 10;
        break;
    }

    p.addBaseData(sb, GPlayerSet(getPlayerId()));
}

/** Starship cost calculator. Allows users to set up build orders on any planet.
    \param planet Planet to use for predictions. May be any planet, or null. */
void
doShipCostCalculator(GPlanet* planet)
{
    if (!planet) {
        // Create a completely new planet.
        TPlanet pl;
        pl.owner     = getPlayerId();
        pl.planet_id = 42;
        pl.fcode[0]  = pl.fcode[1] = pl.fcode[2] = '?';
        pl.mines     = 10;
        pl.factories = 10;
        pl.defense   = 10;
        std::fill_n(pl.ore_mined, countof(pl.ore_mined), 1000);
        pl.colonists = 100;
        pl.supplies  = 10000;
        pl.money     = 10000;
        std::fill_n(pl.ore_ground, countof(pl.ore_ground), 10000);
        std::fill_n(pl.ore_density, countof(pl.ore_density), 50);
        pl.colonist_tax   = 0;
        pl.native_tax     = 0;
        pl.colonist_happy = 100;
        pl.native_happy   = 100;
        pl.native_gov     = 0;
        pl.natives        = 0;
        pl.native_race    = 0;
        pl.temp_code      = 50;
        pl.build_base     = 0;

        GPlanet p(pl.planet_id);
        p.setXY(1000, 1000);
        p.setName("Magrathea");
        p.addPlanetData(pl, GPlayerSet(getPlayerId()));

        createDefaultBase(p);
        doSCC(p, false);
    } else if (planet->isPlayable(GObject::Playable) && planet->hasBase()) {
        // This planet is played by us, and it has a base,
        // so we can use it as-is.
        doSCC(*planet, true);
    } else {
        // Make a copy of the planet
        GPlanet p(*planet);
        fillUnknownFields(p);
        createDefaultBase(p);
        doSCC(p, false);
    }
}
