/**
  *  \file game/selio.cc
  *  \brief Saving and Loading Selection
  *
  *  File format:
  *  - line 1: "CCsel0 <timestamp> <nlayers>"
  *  - line 2: "p<id> <m>" or "s<id> m"
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/selio.h"
#include "cpluslib/format.h"
#include "cpluslib/string.h"
#include "game/global.h"
#include "game/sel.h"
#include "io/textfile.h"
#include "util/answer.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"
#include "game/actions/preconditions.h"

/** Get selection mask value for saving.
    \param slot Selection slot (bit/object number) to query
    \param layer Layer requested by user ([0,NUM_SELECTION_LAYERS), or -1 for all layers) */
static int
getSelectionMask(int slot, int layer)
{
    if (layer < 0) {
        /* All layers */
        int result = 0;
        for (int i = 0; i < GMultiSelection::NUM_SELECTION_LAYERS; ++i) {
            if (GMultiSelection::getSelectionLayer(i).get(slot)) {
                result |= 1 << i;
            }
        }
        return result;
    } else {
        /* One layer */
        return GMultiSelection::getSelectionLayer(layer).get(slot);
    }
}

/** Read selection from file.
    Corresponds roughly to search.cpp::LoadSelectionFromFile.
    \param tf Text file to read from
    \param sel Read into this buffer
    \retval true success
    \retval false file format error */
static bool
readSelection(TextFile& tf, GSelection sel[GMultiSelection::NUM_SELECTION_LAYERS])
{
    string_t line;
    while (tf.getLine(line)) {
        if (line.size() != 0) {
            /* Classify line */
            int slot;
            int limit;
            if (line[0] == '}') {
                return true;
            } else if (line[0] == 'P' || line[0] == 'p') {
                slot = GSelection::FIRST_PLANET_INDEX;
                limit = NUM_PLANETS;
            } else if (line[0] == 'S' || line[0] == 's') {
                slot = GSelection::FIRST_SHIP_INDEX;
                limit = NUM_SHIPS;
            } else {
                return false;
            }

            /* Parse line */
            int id, mask;
            string_t::size_type n = line.find(' ');
            if (n == string_t::npos)
                return false;
            if (!stringValue(line.substr(1, n-1), id) || id <= 0 || id > limit)
                return false;
            if (!stringValue(line.substr(n+1), mask) || mask < 0 || mask > 255)
                return false;

            /* Mark it */
            for (int i = 0; i < GMultiSelection::NUM_SELECTION_LAYERS; ++i)
                if (mask & (1 << i))
                    sel[i].set(slot + id-1, true);
        }
    }
    return true;
}

/**************************** Public Functions ***************************/

/** Save selection.
    \param tf File to write to
    \param layer Layer to save, [0,NUM_SELECTION_LAYERS). -1 to save all layers.
    \param flags selio_Timeless to save a timeless file */
void
saveSelection(TextFile& tf, int layer, int flags)
{
    /* Must have a turn! */
    mustHaveTurn();

    /* Make sure selection is consistent with universe */
    GMultiSelection::getSelectionLayer(GMultiSelection::getCurrentSelectionLayer())
        .copyFrom(getDisplayedTurn().getCurrentUniverse());

    /* Build header */
    string_t header = "CCsel0 ";
    if (flags & selio_Timeless) {
        header += "-";
    } else {
        header += getDisplayedTurn().getCurrentUniverse().getTimestamp().getTimestampAsString();
    }
    if (layer < 0) {
        header += " 8";
    } else {
        header += " 1";
    }
    tf.writeLine(header);

    /* Write file */
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        if (int val = getSelectionMask(GSelection::FIRST_SHIP_INDEX + (sid-1), layer)) {
            tf.writeLine(format("s%d %d", sid, val));
        }
    }
    for (int pid = 1; pid <= NUM_PLANETS; ++pid) {
        if (int val = getSelectionMask(GSelection::FIRST_PLANET_INDEX + (pid-1), layer)) {
            tf.writeLine(format("p%d %d", pid, val));
        }
    }
}

/** Load selection.
    \param tf Text file to read from
    \param layer Current layer, [0,NUM_SELECTION_LAYERS). -1 not allowed here!
    \param flags Acceptance flags. In addition to the selio_Merge flag affecting
                 overall behaviour, this specifies which files are loaded silently.
                 If a file does not match the flags, the AnswerProvider is asked
                 to permit loading anyway. It can also generate an exception.
    \param answer Answer provider, see above. */
void
loadSelection(TextFile& tf, int layer, int flags, AnswerProvider& answer)
{
    int fileFlags = 0;

    /* Must have a turn! */
    mustHaveTurn();

    /* Read header */
    string_t header;
    if (!tf.getLine(header))
        throw FileTooShortException(tf.getName());
    if (strNthWord(header, 0) != "CCsel0")
        throw FileFormatException(tf.getName(), _("File is missing required signature"));

    /* Check layers */
    string_t fileLayer = strNthWord(header, 2);
    if (fileLayer == "1") {
        fileFlags |= selio_AcceptSingle;
    } else if (fileLayer == "8") {
        fileFlags |= selio_AcceptAll;
    } else {
        throw FileFormatException(tf.getName(), _("Invalid layer count"));
    }

    /* Check time */
    string_t fileTime = strNthWord(header, 1);
    if (fileTime != "-" && fileTime != getDisplayedTurn().getCurrentUniverse().getTimestamp().getTimestampAsString()) {
        fileFlags |= selio_Timeless;
    } else {
        fileFlags |= selio_AcceptCurrent;
    }

    /* Check permission */
    if ((flags & fileFlags) != fileFlags) {
        static const char*const questions[] = {
            N_("File contains one selection"),
            N_("File contains all selections"),
            N_("File contains one selection from current turn"),
            N_("File contains all selections from current turn"),
            N_("File contains one selection from a different turn; timestamp is %s"),
            N_("File contains all selections from a different turn; timestamp is %s"),
        };
        int qid = 0;
        if (fileTime != "-") {
            if (fileFlags & selio_AcceptCurrent) {
                qid += 2;
            } else {
                qid += 4;
            }
        }
        if (fileFlags & selio_AcceptAll) {
            qid += 1;
        }
        if (!answer.ask(fileFlags, format(_(questions[qid]), fileTime))) {
            return;
        }
    }

    /* Read data into temporary selection buffer */
    GSelection tmp[GMultiSelection::NUM_SELECTION_LAYERS];
    if (!readSelection(tf, tmp))
        throw FileFormatException(tf.getName(), _("File format error"));

    /* Assimilate into main database */
    int nlayers, firstlayer;
    if (fileFlags & selio_AcceptAll) {
        nlayers    = GMultiSelection::NUM_SELECTION_LAYERS;
        firstlayer = 0;
    } else {
        nlayers    = 1;
        firstlayer = layer;
    }
    for (int L = 0; L < nlayers; ++L) {
        GSelection& sel = GMultiSelection::getSelectionLayer(firstlayer + L);
        if (!(flags & selio_Merge))
            sel.clear();
        for (int slot = 0; slot < GSelection::NUM_WORDS; ++slot)
            sel[slot] |= tmp[L][slot];
        sel.limitToExistingObjects(getDisplayedTurn().getCurrentUniverse());
    }

    /* In any case, this operation has caused the main selection to be changed, so update everything. */
    GMultiSelection::getSelectionLayer(GMultiSelection::getCurrentSelectionLayer())
        .copyTo(getDisplayedTurn().getCurrentUniverse());
    GMultiSelection::sig_selection_changed.raise();
}
