/**
  *  \file game/shippredict.h
  *  \brief Ship Prediction
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SHIPPREDICT_H
#define PCC_V2_GAME_SHIPPREDICT_H

#include "game/struct.h"
#include "game/ship.h"

class GUniverse;

/** Ship Turn Predictor. This object manages prediction of a ship's
    future, especially movement fuel and ETA computation.

    To compute prediction for a single ship, use the two-argument
    constructor. To predict a tower/towee pair, construct a secondary
    object for the towee, and pass it in to the tower's predictor
    using the three-argument constructor. The tower's computeTurn()
    method will then also compute the towee's prediction. */
class GShipTurnPredictor {
 public:
    enum {
        UsedMission = 1,
        UsedFCode = 2,
        UsedShipyard = 4,
        UsedRepair = 8,
        UsedCloak = 16,
        UsedDamageLimit = 32,
        UsedTowee = 64,
        UsedAlchemy = 128
    };

    GShipTurnPredictor(const GUniverse& turn, int id);
    GShipTurnPredictor(const GUniverse& turn, int id, GShipTurnPredictor& towee_override);

    // Fuel usage
    int32_t  getMovementFuelUsed() const;
    int32_t  getCloakFuelUsed() const;

    // Time
    int      getNumTurns() const;
    bool     isAtTurnLimit() const;

    // Computation
    void     computeTurn();
    void     computeMovement();
    int      getUsedProperties() const;

    // Manipulation
    void     setPosition(GPoint pt);
    void     setWaypoint(GPoint pt);
    void     setSpeed(int warp);
    void     setEngine(int engine);
    void     setMission(int m, int i, int t);
    void     setFCode(string_t s);
    void     setFuel(int fuel);

    bool     isAtWaypoint() const;
    GPoint   getPosition() const;
    int      getFuel() const;
    int      getCargo(GCargoType el) const;
    int      getSpeed() const;
    int      getRealOwner() const;
    string_t getFCode() const;
    const GUniverse& getUniverse() const;

    static const int MOVEMENT_TIME_LIMIT = 30;

 private:
    void init();

    int                 id;
    TShip               ship;
    bool                valid;
    GShipTurnPredictor* towee_override;
    const GUniverse&    univ;
    int32_t             movement_fuel_used;
    int32_t             cloak_fuel_used;
    int                 turns_computed;
    int                 used_properties;
};

inline int
GShipTurnPredictor::getUsedProperties() const
{
    return used_properties;
}

/** Get total fuel used for movement. */
inline int32_t
GShipTurnPredictor::getMovementFuelUsed() const
{
    return movement_fuel_used;
}

/** Get total amount of fuel used for cloaking. */
inline int32_t
GShipTurnPredictor::getCloakFuelUsed() const
{
    return cloak_fuel_used;
}

/** Get number of turns computed. */
inline int
GShipTurnPredictor::getNumTurns() const
{
    return turns_computed;
}

/** Check whether computation was stopped because the turn limit was exceeded. */
inline bool
GShipTurnPredictor::isAtTurnLimit() const
{
    return turns_computed >= MOVEMENT_TIME_LIMIT;
}

/** Override this ship's speed. */
inline void
GShipTurnPredictor::setSpeed(int warp)
{
    ship.warp = warp;
}

/** Override this ship's engine. */
inline void
GShipTurnPredictor::setEngine(int engine)
{
    ship.engine = engine;
}

/** Override this ship's mission. */
inline void
GShipTurnPredictor::setMission(int m, int i, int t)
{
    ship.mission           = m;
    ship.mission_intercept = i;
    ship.mission_tow       = t;
}

/** Override this ship's amount of fuel. */
inline void
GShipTurnPredictor::setFuel(int fuel)
{
    ship.ore_cargo[el_Neutronium] = fuel;
}

/** Check whether this ship has reached its waypoint. */
inline bool
GShipTurnPredictor::isAtWaypoint() const
{
    return ship.waypoint_dx == 0 && ship.waypoint_dy == 0;
}

/** Get this ship's current position. */
inline GPoint
GShipTurnPredictor::getPosition() const
{
    return GPoint(ship.x, ship.y);
}

/** Get this ship's current amount of fuel. */
inline int32_t
GShipTurnPredictor::getFuel() const
{
    return ship.ore_cargo[el_Neutronium];
}

/** Get this ship's current speed. */
inline int
GShipTurnPredictor::getSpeed() const
{
    return ship.warp;
}

/** Get this ship's real owner. */
inline int
GShipTurnPredictor::getRealOwner() const
{
    return ship.owner;
}

/** Get the universe used for predicting. */
inline const GUniverse&
GShipTurnPredictor::getUniverse() const
{
    return univ;
}

#endif
