/**
  *  \file u/t_game_mission.cc
  *  \brief Test for game/mission.cc
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/mission.h"

#include "u/t_game.h"
#include "io/memstream.h"

void
GameMissionTestSuite::testMissionIni()
{
    // Generate a pseudo file
    const char data[] = ";22 comment\n"
        "10 one\n"
        "11 two (I:IA)*\n"
        "12 three (T:TA)#\n"
        "13 ~four (I:IA, T:TA)*#\n"
        "14 five(T:TA, I:IA) *#\n"
        "15 six (T:TA I:IA) *#\n"
        "777 seven (whatever) *#\n";
    MemoryStream ms(data);

    // Read it
    GMissionList list;
    list.loadFromIniFile(ms);

    // Must have seven missions
    TS_ASSERT_EQUALS(list.getSize(), 7U);
    TS_ASSERT_EQUALS(list[0].getNumber(), 10);
    TS_ASSERT_EQUALS(list[1].getNumber(), 11);
    TS_ASSERT_EQUALS(list[2].getNumber(), 12);
    TS_ASSERT_EQUALS(list[3].getNumber(), 13);
    TS_ASSERT_EQUALS(list[4].getNumber(), 14);
    TS_ASSERT_EQUALS(list[5].getNumber(), 15);
    TS_ASSERT_EQUALS(list[6].getNumber(), 777);

    // Check names
    TS_ASSERT_EQUALS(list[0].getName(), "one");
    TS_ASSERT_EQUALS(list[1].getName(), "two");
    TS_ASSERT_EQUALS(list[2].getName(), "three");
    TS_ASSERT_EQUALS(list[3].getName(), "~four");     // tilde is not evaluated in mission.ini!
    TS_ASSERT_EQUALS(list[4].getName(), "five");
    TS_ASSERT_EQUALS(list[5].getName(), "six");
    TS_ASSERT_EQUALS(list[6].getName(), "seven");

    // Check presence of Tow parameters
    TS_ASSERT(list[0].getParameterType(GMission::TowParameter) == 0);
    TS_ASSERT(list[1].getParameterType(GMission::TowParameter) == 0);
    TS_ASSERT(list[2].getParameterType(GMission::TowParameter) != 0);
    TS_ASSERT(list[3].getParameterType(GMission::TowParameter) != 0);
    TS_ASSERT(list[4].getParameterType(GMission::TowParameter) != 0);
    TS_ASSERT(list[5].getParameterType(GMission::TowParameter) != 0);
    TS_ASSERT(list[6].getParameterType(GMission::TowParameter) != 0);

    // Check presence of Intercept parameters
    TS_ASSERT(list[0].getParameterType(GMission::InterceptParameter) == 0);
    TS_ASSERT(list[1].getParameterType(GMission::InterceptParameter) != 0);
    TS_ASSERT(list[2].getParameterType(GMission::InterceptParameter) == 0);
    TS_ASSERT(list[3].getParameterType(GMission::InterceptParameter) != 0);
    TS_ASSERT(list[4].getParameterType(GMission::InterceptParameter) != 0);
    TS_ASSERT(list[5].getParameterType(GMission::InterceptParameter) != 0);
    TS_ASSERT(list[6].getParameterType(GMission::InterceptParameter) != 0);

    // Check names of Tow parameters
    TS_ASSERT_EQUALS(list[2].getParameterName(GMission::TowParameter), "TA");
    TS_ASSERT_EQUALS(list[3].getParameterName(GMission::TowParameter), "TA");
    TS_ASSERT_EQUALS(list[4].getParameterName(GMission::TowParameter), "TA");
    TS_ASSERT_EQUALS(list[5].getParameterName(GMission::TowParameter), "TA");
    TS_ASSERT_EQUALS(list[6].getParameterName(GMission::TowParameter), "Tow");

    // Check names of Intercept parameters
    TS_ASSERT_EQUALS(list[1].getParameterName(GMission::InterceptParameter), "IA");
    TS_ASSERT_EQUALS(list[3].getParameterName(GMission::InterceptParameter), "IA");
    TS_ASSERT_EQUALS(list[4].getParameterName(GMission::InterceptParameter), "IA");
    TS_ASSERT_EQUALS(list[5].getParameterName(GMission::InterceptParameter), "IA");
    TS_ASSERT_EQUALS(list[6].getParameterName(GMission::InterceptParameter), "Intercept");

    // Check hotkeys. Hotkeys are assigned in sequential order.
    TS_ASSERT_EQUALS(list[0].getHotkey(), 'a');
    TS_ASSERT_EQUALS(list[1].getHotkey(), 'b');
    TS_ASSERT_EQUALS(list[2].getHotkey(), 'c');
    TS_ASSERT_EQUALS(list[3].getHotkey(), 'd');
    TS_ASSERT_EQUALS(list[4].getHotkey(), 'e');
    TS_ASSERT_EQUALS(list[5].getHotkey(), 'f');
    TS_ASSERT_EQUALS(list[6].getHotkey(), 'g');
}

void
GameMissionTestSuite::testMissionIniRaces()
{
    // Generate a pseudo file
    const char data[] =
        "10 one/1\n"
        "11 two/2 (I:IA)*\n"
        "12 three (T:TA)#/3\n"
        "13 four (I:IA, T:TA)  *# /4\n"
        "14 fi/ve(T:TA2, I:IA1) *#\n"
        "15 six/1A (T:TA I:IA) *#\n"
        "16 seven (whatever) *# /2B\n";
    MemoryStream ms(data);

    // Read it
    GMissionList list;
    list.loadFromIniFile(ms);

    // Must have seven missions
    TS_ASSERT_EQUALS(list.getSize(), 7U);
    for (int i = 0; i < 7; ++i) {
        TS_ASSERT_EQUALS(list[i].getNumber(), 10+i);
    }

    // Check names
    TS_ASSERT_EQUALS(list[0].getName(), "one");
    TS_ASSERT_EQUALS(list[1].getName(), "two");
    TS_ASSERT_EQUALS(list[2].getName(), "three");
    TS_ASSERT_EQUALS(list[3].getName(), "four");
    TS_ASSERT_EQUALS(list[4].getName(), "fi/ve");
    TS_ASSERT_EQUALS(list[5].getName(), "six");
    TS_ASSERT_EQUALS(list[6].getName(), "seven");

    // Check presence of Tow parameters
    TS_ASSERT(list[0].getParameterType(GMission::TowParameter) == 0);
    TS_ASSERT(list[1].getParameterType(GMission::TowParameter) == 0);
    TS_ASSERT(list[2].getParameterType(GMission::TowParameter) != 0);
    TS_ASSERT(list[3].getParameterType(GMission::TowParameter) != 0);
    TS_ASSERT(list[4].getParameterType(GMission::TowParameter) != 0);
    TS_ASSERT(list[5].getParameterType(GMission::TowParameter) != 0);
    TS_ASSERT(list[6].getParameterType(GMission::TowParameter) != 0);

    // Check presence of Intercept parameters
    TS_ASSERT(list[0].getParameterType(GMission::InterceptParameter) == 0);
    TS_ASSERT(list[1].getParameterType(GMission::InterceptParameter) != 0);
    TS_ASSERT(list[2].getParameterType(GMission::InterceptParameter) == 0);
    TS_ASSERT(list[3].getParameterType(GMission::InterceptParameter) != 0);
    TS_ASSERT(list[4].getParameterType(GMission::InterceptParameter) != 0);
    TS_ASSERT(list[5].getParameterType(GMission::InterceptParameter) != 0);
    TS_ASSERT(list[6].getParameterType(GMission::InterceptParameter) != 0);

    // Check names of Tow parameters
    TS_ASSERT_EQUALS(list[2].getParameterName(GMission::TowParameter), "TA");
    TS_ASSERT_EQUALS(list[3].getParameterName(GMission::TowParameter), "TA");
    TS_ASSERT_EQUALS(list[4].getParameterName(GMission::TowParameter), "TA2");
    TS_ASSERT_EQUALS(list[5].getParameterName(GMission::TowParameter), "TA");
    TS_ASSERT_EQUALS(list[6].getParameterName(GMission::TowParameter), "Tow");

    // Check names of Intercept parameters
    TS_ASSERT_EQUALS(list[1].getParameterName(GMission::InterceptParameter), "IA");
    TS_ASSERT_EQUALS(list[3].getParameterName(GMission::InterceptParameter), "IA");
    TS_ASSERT_EQUALS(list[4].getParameterName(GMission::InterceptParameter), "IA1");
    TS_ASSERT_EQUALS(list[5].getParameterName(GMission::InterceptParameter), "IA");
    TS_ASSERT_EQUALS(list[6].getParameterName(GMission::InterceptParameter), "Intercept");

    // Check race masks
    TS_ASSERT_EQUALS((list[0].getRaceMask().toInteger() & 0xFFE), 0x002U);
    TS_ASSERT_EQUALS((list[1].getRaceMask().toInteger() & 0xFFE), 0x004U);
    TS_ASSERT_EQUALS((list[2].getRaceMask().toInteger() & 0xFFE), 0x008U);
    TS_ASSERT_EQUALS((list[3].getRaceMask().toInteger() & 0xFFE), 0x010U);
    TS_ASSERT_EQUALS((list[4].getRaceMask().toInteger() & 0xFFE), 0xFFEU);
    TS_ASSERT_EQUALS((list[5].getRaceMask().toInteger() & 0xFFE), 0x402U);
    TS_ASSERT_EQUALS((list[6].getRaceMask().toInteger() & 0xFFE), 0x804U);
}
